# Copyright 2018 Comcast Cable Communications Management, LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines a single worker to collect stats from a single host"""

import multiprocessing
import math
from . import utils, config, ping, traceroute, ports

def dummy(_):
	pass

class Collector(multiprocessing.Process):
	"""
	A threaded worker that collects stats for a single host.
	"""
	trace = None
	result = [utils.PingResult(-1, -1, -1, -1, 100.),
	          utils.Trace([utils.TraceStep('*', -1)] * 10),
	          utils.ScanResult(None, None, None)]

	def __init__(self, host:str, ID:int, conf:config.Config = config.CONFIG):
		"""
		Initializes the Collector, and its worker pool
		"""
		super(Collector, self).__init__()

		self.hostname = host
		self.conf = conf
		self.host = conf.HOSTS[host]
		self.name = host
		self.ID = ID

		self.pipe = multiprocessing.Pipe()

	def run(self):
		"""
		Called when the thread is run
		"""
		with multiprocessing.pool.ThreadPool() as pool:
			pscan_result, trace_result, ping_result = None, None, None
			if self.conf.PORTSCAN:
				pscan_result = pool.apply_async(ports.portScan,
													 (self.host, pool),
													 error_callback=utils.error)
			if self.conf.TRACE:
				trace_result = pool.apply_async(traceroute.trace,
													 (self.host, self.ID, self.conf),
													 error_callback=utils.error)
			if not self.conf.NOPING:
				try:
					self.ping(pool)
				except (multiprocessing.TimeoutError, ValueError):
					self.result[0] = type(self).result[0]
			else:
				self.result[0] = None

			if self.conf.TRACE:
				try:
					self.result[1] = trace_result.get(self.conf.HOPS)
				except multiprocessing.TimeoutError:
					self.result[1] = type(self).result[1]
			else:
				self.result[1] = None

			if self.conf.PORTSCAN:
				try:
					self.result[2] = pscan_result.get(0.5)
				except multiprocessing.TimeoutError:
					self.result[2] = type(self).result[2]
			else:
				self.result[2] = None
			try:
				self.pipe[1].send(self.result)
			except OSError as e:
				utils.error(OSError("Error sending results: %s" % e))

	def ping(self, pool:multiprocessing.pool.ThreadPool, pinger:ping.Pinger = None):
		"""
		Pings the host
		"""
		destroyPinger = dummy
		if pinger is None:
			pinger = ping.Pinger(self.host, bytes(self.conf.PAYLOAD))
			destroyPinger = lambda x: x.sock.close()

		# Aggregates round-trip time for each packet in the sequence
		rtt, lost = [], 0

		# Sends, receives and parses all icmp packets asynchronously
		results = pool.map_async(pinger.ping,
		                              range(self.conf.NUMPINGS),
		                              error_callback=utils.error)
		pkts = results.get(2)

		for pkt in pkts:
			if pkt != None and pkt > 0:
				rtt.append(pkt*1000)
			else:
				lost += 1

		try:
			avg = sum(rtt) / len(rtt)
			std = 0.
			for item in rtt:
				std += (avg - item)**2
			std /= len(rtt) - 1
			std = math.sqrt(std)
		except ZeroDivisionError:
			std = 0.
		finally:
			destroyPinger(pinger)

		if rtt:
			self.result[0] = utils.PingResult(min(rtt), avg, max(rtt), std, lost/self.conf.NUMPINGS *100.0)
		else:
			self.result[0] = type(self).result[0]


	def __str__(self) -> str:
		"""
		Implements 'str(self)'

		Returns a plaintext output result
		"""
		ret = []
		if self.host[0] == self.hostname:
			ret.append(self.hostname)
		else:
			ret.append("%s (%s)" % (self.hostname, self.host[0]))

		pings, trace, scans = self.result

		if pings:
			ret.append(str(pings))
		if trace and trace != self.trace:
			self.trace = trace
			# Dirty hack because I can't inherit with strong typing in Python 3.4
			ret.append(utils.traceToStr(trace))
		if scans:
			ret.append(str(scans))

		return "\n".join(ret)

	def __repr__(self) -> repr:
		"""
		Implements `repr(self)`

		Returns a JSON output result
		"""
		ret = [r'{"addr":"%s"' % self.host[0]]
		ret.append(r'"name":"%s"' % self.hostname)

		if not self.conf.NOPING:
			ret.append(r'"ping":%s' % repr(self.result[0]))

		if self.conf.TRACE and self.trace != self.result[1]:
			self.trace = self.result[1]
			# Dirty hack because I can't inherit with strong typing in Python 3.4
			ret.append(r'"trace":%s' % utils.traceRepr(self.result[1]))

		if self.conf.PORTSCAN:
			ret.append(r'"scan":%s' % repr(self.result[2]))

		return ','.join(ret) + '}'


	def recv(self):
		"""
		Returns a message from the Collector's Pipe
		"""
		return self.pipe[0].recv()



# Copyright 2018 Comcast Cable Communications Management, LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module defines a class and utilities to manage icmp/icmpv6 echo requests
and replies to remote hosts.
"""

import socket
import struct
import time
import sys
import math
from . import utils
from . import icmp

def icmpParse(pkt: bytes, ipv6: bool) -> int:
	"""
	Parses an icmp packet, returning its sequence number.

	If the packet is found to be not an echo reply, this will
		immediately return -1, indicating that this packet
		should be disregarded.
	"""
	try:
		if ipv6:
			if pkt[0] == 129:
				return struct.unpack("!H", pkt[6:8])[0]
			return -1
		if pkt[20] == 0:
			return struct.unpack("!H", pkt[26:28])[0]
		return -1
	except (IndexError, struct.error):
		return -1

class Pinger():
	"""
	A data structure that handles icmp pings to a remote machine.
	"""
	def __init__(self, host: utils.Host, payload: bytes):
		"""
		Inializes a socket connection to the host on port 22, and returns a Pinger object
		referencing it.
		"""

		self.sock, self.icmpParse, self.mkPkt = None, None, None

		if host.family == socket.AF_INET6:
			self.sock = socket.socket(host.family, socket.SOCK_RAW, proto=socket.IPPROTO_ICMPV6)
			self.icmpParse = self._icmpv6Parse
			self.mkPkt = self._mkPkt6
		else:
			self.sock = socket.socket(host.family, socket.SOCK_RAW, proto=socket.IPPROTO_ICMP)
			self.icmpParse = self._icmpv4Parse
			self.mkPkt = self._mkPkt4

		self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

		self.sock.settimeout(1)
		self.payload = payload

		#Build a socket object
		self.host = host

		self.timestamps = {}

	def sendAll(self, num:int) -> utils.PingResult:
		"""
		Sends all pings sequentially in one thread
		"""
		for i in range(num):
			pkt = icmp.ICMPPkt(self.host, payload=struct.pack("!HH", 2, i))
			self.timestamps[i] = time.time()

			try:
				self.sock.sendto(bytes(pkt), (self.host.addr, 0))
			except Exception as e:
				utils.err("Network is unreachable... (%s)" % e)

		return self.recvAll(num)

	def recvAll(self, num:int) -> utils.PingResult:
		"""
		Recieves and parses all packets
		"""
		maxPacketLen, found = 100 + len(self.payload), 0
		pkts = [-1,]*num

		while found < num:

			try:
				pkt, addr = self.sock.recvfrom(maxPacketLen)
			except (socket.timeout, TimeoutError):
				# If we hit a socket timeout, we'll consider the packet dropped
				found += 1
				continue

			if addr[0] == self.host[0]:
				seqno = self.icmpParse(pkt)
				if seqno in range(len(pkts)):
					pkts[seqno] = time.time() - self.timestamps[seqno]
					found += 1

		# All packets collected; parse and return the results
		rtt, lost = [], 0
		for pkt in pkts:
			if pkt > 0:
				rtt.append(pkt*1000)
			else:
				lost += 1

		if lost == num:
			return utils.PingResult(-1, -1, -1, -1, 100.)

		try:
			avg = sum(rtt) / len(rtt)
			std = 0.
			for item in rtt:
				std += (avg - item) ** 2
			std /= len(rtt) - 1
			std = math.sqrt(std)
		except ZeroDivisionError:
			std = 0.

		return utils.PingResult(min(rtt), avg, max(rtt), std, lost/num * 100.0)

	def ping(self, seqno: int) -> float:
		"""
		Sends a single icmp packet to the remote host.
		Returns the round-trip time (in ms) between packet send and receipt
		or 0 if packet was not received.
		"""
		pkt = icmp.ICMPPkt(self.host, payload=struct.pack("!HH", 2, seqno))

		# I set time here so that rtt includes the device latency
		self.timestamps[seqno] = time.time()

		try:
			# ICMP has no notion of port numbers
			self.sock.sendto(bytes(pkt), (self.host.addr, 0))
		except Exception as e:
			#Sometimes, when the network is unreachable this will erroneously report that there's an
			#'invalid argument', which is impossible since the hostnames are coming straight from
			#`socket` itself
			raise Exception("Network is unreachable... (%s)" % e)
		return self.recv()

	@staticmethod
	def _icmpv4Parse(pkt: bytes) -> int:
		"""
		Attemtps to parse an icmpv4 packet, returning the sequence number if parsing succeds,
		or -1 otherwise.
		"""
		try:
			if pkt[20] == 0:
				return struct.unpack("!H", pkt[26:28])[0]
		except (IndexError, struct.error):
			pass
		return -1

	@staticmethod
	def _icmpv6Parse(pkt: bytes) -> int:
		"""
		Attemtps to parse an icmpv6 packet, returning the sequence number if parsing succeds,
		or -1 otherwise.
		"""
		try:
			if pkt[0] == 0x81:
				return struct.unpack("!H", pkt[6:8])[0]
		except (IndexError, struct.error):
			pass
		return -1

	def _mkPkt4(self, seqno: int) -> bytes:
		"""
		Contsructs and returns an ICMPv4 packet
		"""
		header = struct.pack("!BBHHH", 8, 0, 0, 2, seqno)
		checksum = self._checksum4(header + self.payload)
		return struct.pack("!BBHHH", 8, 0, checksum, 2, seqno) + self.payload

	def _mkPkt6(self, seqno: int) -> bytes:
		"""
		Contsructs and returns an ICMPv6 packet
		"""
		header = struct.pack("!BBHHH", 0x80, 0, 0, 2, seqno)
		checksum = self._checksum6(header)
		return struct.pack("!BBHHH", 0x80, 0, checksum, 2, seqno) + self.payload

	@staticmethod
	def _checksum4(pkt: bytes) -> int:
		"""
		calculates and returns the icmpv4 checksum of 'pkt'
		"""

		countTo = len(pkt) // 2 * 2
		total, count = 0, 0

		# Handle bytes in pairs (decoding as short ints)
		loByte, hiByte = 0, 0
		while count < countTo:
			if sys.byteorder == "little":
				loByte = pkt[count]
				hiByte = pkt[count + 1]
			else:
				loByte = pkt[count + 1]
				hiByte = pkt[count]
			total += hiByte * 256 + loByte
			count += 2

		# Handle last byte if applicable (odd-number of bytes)
		# Endianness should be irrelevant in this case
		if countTo < len(pkt): # Check for odd length
			total += pkt[len(pkt) - 1]

		total &= 0xffffffff # Truncate sum to 32 bits (a variance from ping.c, which
		                    # uses signed ints, but overflow is unlikely in ping)

		total = (total >> 16) + (total & 0xffff)    # Add high 16 bits to low 16 bits
		total += (total >> 16)                      # Add carry from above (if any)

		return socket.htons((~total) & 0xffff)

	def _checksum6(self, pkt: bytes) -> int:
		"""
		calculates and returns the icmpv6 checksum of pkt
		"""
		laddr = socket.inet_pton(self.host[1], self.sock.getsockname()[0])
		raddr = socket.inet_pton(*reversed(self.host))
		# IPv6 Pseudo-Header used for checksum calculation as specified by
		# RFC 2460 (https://tools.ieft.org/html/rfc2460)
		psh = laddr + raddr + struct.pack("!I", len(pkt)) + b'\x00\x00\x00:'
		# This last bit is the 4-byte-packed icmp6 protocol number (58 or 0xa3)


		total, packet = 0, psh+pkt

		# Sum all 2-byte words
		num_words = len(packet) // 2
		for chunk in struct.unpack("!%sH" % num_words, packet[0:num_words*2]):
			total += chunk

		# Add any left-over byte (for odd-length packets)
		if len(packet) % 2:
			total += ord(packet[-1]) << 8

		# Fold 32-bits into 16-bits
		total = (total >> 16) + (total & 0xffff)
		total += total >> 16
		return ~total + 0x10000 & 0xffff

	def recv(self) -> float:
		"""
		Recieves each ping sent.
		"""
		# If a packet is not an echo reply, icmpParse will give its seqno as -1
		# This lets us disregard packets from traceroutes immediately
		maxlen = 100+len(self.payload) #this should be enough to cover headers and whatnot

		while True:

			try:
				pkt, addr = self.sock.recvfrom(maxlen)
			except (socket.timeout, TimeoutError):
				return -1

			# The packet must have actually come from the host we pinged
			if addr[0] == self.host[0]:
				seqno = self.icmpParse(pkt)
				if seqno >= 0:
					return time.time() - self.timestamps[seqno]


	def __enter__(self) -> "Pinger":
		"""
		Context-managed instantiation
		"""
		return self

	def __exit__(self, exc_type, exc_value, traceback):
		"""
		Context-managed cleanup
		"""
		self.sock.close()

		if exc_type and exc_value:
			utils.error(exc_type("Unknown error occurred while pinging"))
			utils.error(exc_type(exc_value))
			if traceback:
				utils.warn(traceback)



"""
This module contains convenience functions for dealing with the ICMP protocol.
Most notably, it contains the `ICMPPkt` class, which can construct an ICMP packet object either
from a simple payload and destination (which will make an 'Echo Request' of the appropriate version)
or from a raw bytestring that is presumably an ICMP packet recived from an external host.
"""

# Copyright 2018 Comcast Cable Communications Management, LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
import struct
import enum
from . import utils



# Gets our local IP address, for calculating ICMPv6 checksums
with socket.socket(socket.AF_INET6, socket.SOCK_DGRAM) as s:
	s.connect(("2001:4998:c:1023::4", 1)) #yahoo.com public IPv6 address
	LADDR = socket.inet_pton(s.family, s.getsockname()[0])

LADDR = LADDR # 'LADDR' is now a global value

# Note that these type code mappings only enumerate the types used by connvitals
class ICMPType(enum.IntEnum):
	"""
	Mapping of ICMP type codes to their names.

	Meant only to be inherited for it's `str` type coercion behaviour, IPv4 and IPv6 have their own,
	more specific Type enumerations.
	"""
	def __str__(self) -> str:
		"""
		Gives the name of the ICMP Type.
		"""
		return self.name.replace('_', ' ')

class ICMPv4Type(ICMPType):
	"""
	Mapping of ICMPv4 type codes to their names.
	"""
	Echo_Reply = 0
	Destination_Unreachable = 3
	Echo_Request = 8
	Time_Exceeded = 11

class ICMPv6Type(ICMPType):
	"""
	Mapping of ICMPv6 type codes to their names.
	"""
	Destination_Unreachable = 1
	Time_Exceeded = 3
	Echo_Request = 128
	Echo_Reply = 129

def ICMPv4_checksum(pkt: bytes) -> int:
	"""
	Implementation of the "Internet Checksum" specified in
	RFC 1071 (https://tools.ieft.org/html/rfc1071)

	Ideally this would act on the string as a series of half-words in host byte order,
	but this works.

	Network data is big-endian, hosts are typically little-endian,
	which makes this much more tedious than it needs to be.
	"""

	from sys import byteorder

	countTo = len(pkt) // 2 * 2
	total, count = 0, 0

	# Handle bytes in pairs (decoding as short ints)
	loByte, hiByte = 0, 0
	while count < countTo:
		if byteorder == "little":
			loByte = pkt[count]
			hiByte = pkt[count + 1]
		else:
			loByte = pkt[count + 1]
			hiByte = pkt[count]
		total += hiByte * 256 + loByte
		count += 2

	# Handle last byte if applicable (odd-number of bytes)
	# Endianness should be irrelevant in this case
	if countTo < len(pkt): # Check for odd length
		total += pkt[len(pkt) - 1]

	total &= 0xffffffff # Truncate sum to 32 bits (a variance from ping.c, which
	                    # uses signed ints, but overflow is unlikely in ping)

	total = (total >> 16) + (total & 0xffff)    # Add high 16 bits to low 16 bits
	total += (total >> 16)                      # Add carry from above (if any)

	return socket.htons((~total) & 0xffff)

def ICMPv6_checksum(pkt: bytes, laddr: bytes, raddr: bytes) -> int:
	"""
	Implementation of the ICMPv6 "Internet Checksum" as specified in
	RFC 1701 (https://tools.ieft.org/html/rfc1701).

	This takes the Payload Length from the IPv6 layer to be 32 (0x20), since we
	don't expect any extension headers and ICMP doesn't carry any length
	information.
		pkt: A complete ICMP packet, with the checksum field set to 0
		laddr: The (fully-expanded) local address of the socket that will send pkt
		raddr: The (fully-expanded) remote address of the host to which the pkt will be sent
		returns: A bytes object representing the checksum
	"""

	# IPv6 Pseudo-Header used for checksum calculation as specified by
	# RFC 2460 (https://tools.ieft.org/html/rfc2460)
	psh = laddr + raddr + struct.pack("!I", len(pkt)) + b'\x00\x00\x00:'
	# This last bit is the 4-byte-packed icmp6 protocol number (58 or 0xa3)


	total, packet = 0, psh+pkt

	# Sum all 2-byte words
	num_words = len(packet) // 2
	for chunk in struct.unpack("!%sH" % num_words, packet[0:num_words*2]):
		total += chunk

	# Add any left-over byte (for odd-length packets)
	if len(packet) % 2:
		total += packet[-1] << 8

	# Fold 32-bits into 16-bits
	total = (total >> 16) + (total & 0xffff)
	total += total >> 16
	return ~total + 0x10000 & 0xffff

def ICMP_checksum(pkt: bytes, raddr: bytes = None) -> int:
	"""
	This is an abstraction used to allow calculation of a checksum to be
	agnostic of the protocol version in use.
	"""
	global LADDR
	return ICMPv6_checksum(pkt, LADDR, raddr) if raddr else ICMPv4_checksum(pkt)

class ICMPPkt():
	"""
	This object represents an ICMP packet, with associated convenience functions.
	"""

	fmt = "!BBH4s"
	version = 4 # common default

	def __init__(self, host:utils.Host, pkt:bytes = None, payload:bytes = None):
		"""
		Initializes an ICMP packet, either from raw bytes or a desired host and payload

		If constructed from a host/payload, will set the checksum and construct a 'ping'
		packet.
		"""
		if pkt:
			packetlen = len(pkt)
			if packetlen < 8:
				raise ValueError("Byte string %r too small to be ICMP packet!" % (pkt,))
			elif packetlen > 8:
				utils.warn("ICMP packet may not be valid (has length %d, expected 8)" % (packetlen,))
				pkt = pkt[:8]

			self.outbound = False
			self.Host = host
			Type, self.Code, self.Checksum, self.Payload = struct.unpack(self.fmt, pkt)

			if self.Host.family == socket.AF_INET6:
				self.version = 6
				self.Type = ICMPv6Type(Type)
			else:
				self.Type = ICMPv4Type(Type)

		elif payload:
			self.Host = host
			self.outbound = True
			self.fmt = self.fmt.replace('4', str(4+len(payload)))

			if self.Host[1] == socket.AF_INET6:
				self.version = 6
				self.Type = ICMPv6Type.Echo_Request
			else:
				self.Type = ICMPv4Type.Echo_Request

			self.Code = 0
			self.Payload = payload
			self.Checksum = self.calcChecksum()

		else:
			raise TypeError("ICMPPkt() must be called with a host, and either a packet or payload!")

	def calcChecksum(self) -> int:
		"""
		Calculates the checksum of this ICMP Packet
		"""
		global LADDR

		pkt = struct.pack(self.fmt.replace('H', "2s"), self.Type, self.Code, b'\x00\x00\x00\x00', self.Payload)

		if self.version == 6:

			hostaddr = socket.inet_pton(self.Host.family, self.Host.addr)

			# packet was outbound, proceed as normal
			if self.outbound:
				return ICMP_checksum(pkt, hostaddr)

			# packet was inbound, LADDR and raddr are reversed
			return ICMPv6_checksum(pkt, hostaddr, LADDR)

		return ICMP_checksum(pkt)


	def __bytes__(self) -> bytes:
		"""
		Implements `bytes(self)`

		This builds the packet for sending along a socket.
		"""
		return struct.pack(self.fmt, self.Type, self.Code, self.Checksum, self.Payload)

	def __bool__(self) -> bool:
		"""
		Implements `bool(self)`

		Checks that the `Checksum` attribute matches the calculated checksum.
		"""
		return self.calcChecksum() == self.Checksum

	def __str__(self) -> str:
		"""
		Implements `str(self)`
		"""
		return "%s ICMPv%d packet, bound for %s" % (self.Type, self.version, self.Host.addr)

	def __repr__(self) -> str:
		"""
		Implements `repr(self)`
		"""
		return "ICMPPkt(Type=%r, Code=%d, Payload=%r, Host=%r)" % (self.Type, self.Code, self.Payload, self.Host)

	@property
	def seqno(self) -> int:
		"""
		Gives the sequence number if this is an Echo Request/Reply packet,
		else raises an AttributeError.
		"""
		if self.Type not in {ICMPv6Type.Echo_Reply,
		                     ICMPv6Type.Echo_Request,
		                     ICMPv4Type.Echo_Reply,
		                     ICMPv4Type.Echo_Request}:
			raise AttributeError("Only Echo Requests/Replies have seqno")

		return struct.unpack("!HH", self.Payload)[1]




# Copyright 2018 Comcast Cable Communications Management, LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains functions for scanning some specific hosts for specific
information. Currently has functionality for http(s) servers on ports 80/443
and MySQL servers on port 3306.
"""

import socket
import time
import multiprocessing.pool
import typing
import ssl
from . import utils

class Scanner():
	"""
	Holds persistent information that can be used to repeatedly
	do port scans.
	"""

	# Message constant - so it doesn't need to be re-allocated at runtime.
	HTTP_MSG = b'HEAD / HTTP/1.1\r\nConnection: Keep-Alive\r\nHost: '


	def __init__(self, host:utils.Host):
		"""
		Sets up this scanner, including opening three sockets for
		scanning.
		"""

		self.host = host
		self.HTTP_MSG += self.host.addr.encode() + b'\r\n\r\n'

		# Sets up two TCP sockets, each with a dedicated message
		# buffer - 1 to check the http port, and 1 to check the
		# https port.
		self.buffers = [bytearray(1024), bytearray(1024)]
		self.socks = [
		                 socket.socket(family=host.family),
		                 ssl.wrap_socket(socket.socket(family=host.family), ssl_version=3),
		             ]

		for sock in self.socks:
			sock.settimeout(0.08)


		# When HTTP(S) connections fail, close them immediately. The individual
		# scanners will re-attempt the connection (mysql server connections)
		# shouldn't persist, because there's no mysql 'NOOP' to my knowledge).
		try:
			self.socks[0].connect((self.host.addr, 80))
		except (ConnectionRefusedError, socket.timeout, socket.gaierror):
			utils.warn("Connection Refused by %s on port 80" % self.host.addr)
			self.socks[0].close()

		try:
			self.socks[1].connect((self.host.addr, 443))
		except ssl.SSLError as e:
			utils.warn("SSL handshake with %s failed: %s" % (url[0], e))
			self.socks[1].close()
		except (ConnectionRefusedError, socket.timeout, socket.gaierror):
			utils.warn("Connection Refused by %s on port 443" % self.host.addr)
			self.socks[1].close()


	def __enter__(self) -> 'Scanner':
		"""
		Context-managed `Scanner` object.
		"""

		return self

	def __exit__(self, exc_type, exc_value, traceback):
		"""
		Context-cleanup for `Scanner` object.
		"""
		for sock in self.socks:
			sock.shutdown(socket.SHUT_RDWR)
			sock.close()

		if exc_type and exc_value:
			utils.error("Unknown error occurred (Traceback: %s)" % traceback)
			utils.error(exc_type(exc_value), True)

	def __del__(self):
		"""
		Non-context-managed cleanup.
		"""
		try:
			for sock in self.socks:
				sock.close()
		except AttributeError:
			# Context-management handled the sockets already
			pass


	def scan(self, pool:typing.Union[multiprocessing.pool.Pool, bool] = None) -> utils.ScanResult:
		"""
		Performs a full portscan of the host, and returns a format-able result.

		If the `pool` argument is given, it should be either a usable
		`multiprocessing.pool.Pool` ancestor (i.e. ThreadPool or Pool), or a
		boolean. If `pool` is literally `True`, a new ThreadPool will be
		created to perform the scan.

		If `pool` is `None`, each scan is done sequentially.
		"""
		if pool:
			if pool is True:
				with multiprocessing.pool.ThreadPool(3) as p:
					httpresult = p.apply_async(self.http, ())
					httpsresult = p.apply_async(self.https, ())
					mysqlresult = p.apply_async(self.mysql, ())

					return utils.ScanResult(httpresult.get(), httpsresult.get(), mysqlresult.get())

			httpresult = pool.apply_async(self.http, ())
			httpsresult = pool.apply_async(self.https, ())
			mysqlresult = pool.apply_async(self.mysql, ())

			return utils.ScanResult(httpresult.get(), httpsresult.get(), mysqlresult.get())

		return utils.ScanResult(self.http(), self.https(), self.mysql())

	def http(self) -> typing.Optional[typing.Tuple[float, str, str]]:
		"""
		Checks for http content on port 80.

		This uses a HEAD / HTTP/1.1 request, and returns a tuple containing
		the total latency, the server's status code and reason, and the server's
		name/version (if found in the first kilobyte of data).
		"""

		s = self.socks[0]

		try:
			rtt = time.time()
			s.send(self.HTTP_MSG)
			_ = s.recv_into(self.buffers[0])
			rtt = time.time() - rtt

		except OSError:
			# Possibly the connection was closed; try to re-open.
			try:
				self.socks[0].close()
				self.socks[0] = socket.socket(family=self.host.family)
				self.socks[0].settimeout(0.08)
				self.socks[0].connect((self.host.addr, 80))
				rtt = time.time()
				self.socks[0].send(self.HTTP_MSG)
				_ = self.socks[0].recv_into(self.buffers[0])
				rtt = time.time() - rtt

			except (OSError, socket.gaierror, socket.timeout) as e:
				# If this happens, the server likely went down
				utils.warn("Could not connect to %s:80 - %s" % (self.host.addr, e))
				return None

		except (socket.gaierror, socket.timeout) as e:
			utils.warn("Could not connect to %s:80 - %s" % (self.host.addr, e))
			return None

		if not self.buffers[0]:
			return None

		status = self.buffers[0][9:12].decode()

		try:
			srv = self.buffers[0].index(b'Server: ')
			srv = self.buffers[0][srv+8:self.buffers[0].index(b'\r', srv)].decode()
		except ValueError:
			# Server header not found
			return rtt*1000, status, "Unkown"
		else:
			return rtt*1000, status, srv
		finally:
			# Creating a new buffer is faster than clearing the old one
			self.buffers[0] = bytearray(1024)

	def https(self) -> typing.Optional[typing.Tuple[float, str, str]]:
		"""
		Checks for http content on port 433.

		This uses a HEAD / HTTP/1.1 request, and returns a tuple containing
		the total latency, the server's status code and reason, and the server's
		name/version (if found in the first kilobyte of data).

		Note that this is principally the same as `self.http`, but in the interest
		of favoring time optimization (no conditional branching, fewer dereferences)
		over space optimization the process is repeated nearly verbatim.
		"""

		s = self.socks[1]
		try:
			rtt = time.time()
			s.send(self.HTTP_MSG)
			_ = s.recv_into(self.buffers[1])
			rtt = time.time() - rtt
		except ssl.SSLError as e:
			utils.warn("SSL handshake with %s failed: %s" % (url[0], e))
			return None

		except OSError:
			# Possibly the connection was closed; try to re-open.
			try:
				self.socks[1].close()
				self.socks[1] = ssl.wrap_socket(socket.socket(family=self.host.family), ssl_version=3)
				self.socks[1].settimeout(0.08)
				self.socks[1].connect((self.host.addr, 443))
				rtt = time.time()
				self.socks[1].send(self.HTTP_MSG)
				_ = self.socks[1].recv_into(self.buffers[1])
				rtt = time.time() - rtt
			except ssl.SSLError as e:
				utils.warn("SSL handshake with %s failed: %s" % (url[0], e))
				return None

			except (OSError, socket.gaierror, socket.timeout) as e:
				# If this happens, the server likely went down
				utils.warn("Could not connect to %s:443 - %s" % (self.host.addr, e))
				return None

		except (socket.gaierror, socket.timeout) as e:
			utils.warn("Could not connect to %s:443 - %s" % (self.host.addr, e))
			return None

		if not self.buffers[1]:
			return None

		status = self.buffers[1][9:12].decode()

		try:
			srv = self.buffers[1].index(b'Server: ')
			srv = self.buffers[1][srv+8:self.buffers[1].index(b'\r', srv)].decode()
		except ValueError:
			# Server header not found
			return rtt*1000, status, "Unkown"
		else:
			return rtt*1000, status, srv
		finally:
			# Creating a new buffer is faster than clearing the old one
			self.buffers[1] = bytearray(1024)

	def mysql(self) -> typing.Optional[typing.Tuple[float, str]]:
		"""
		Checks for a MySQL server running on port 3306.

		Returns a tuple containing the total latency and the server version if one is found.
		"""

		with socket.socket(family=self.host.family) as s:

			try:
				rtt = time.time()
				s.connect((self.host.addr, 3306))
				_ = s.recv_into(self.buffers[2])
				ret = sock.recv(1024)

			except (OSError, socket.gaierror, socket.timeout) as e:
				utils.warn("Could not connect to %s:3306 - %s" % (self.host.addr, e))
				return None

		rtt = (time.time() - rtt)*1000
		try:
			return rtt, ret[5:10].decode()
		except UnicodeError:
			utils.warn("Server at %s:3306 doesn't appear to be mysql." % self.host.addr)
			return rtt, "Unknown"


# Functional implementation provided for convenience/legacy support


def http(url: utils.Host, port: int=80) -> typing.Optional[typing.Tuple[float, str, str]]:
	"""
	Checks for http content being served by url on a port passed in ssl.
		(If ssl is 443, wraps the socket with ssl to communicate HTTPS)
	Returns a HEAD request's status code if a server is found, else None
	"""

	# Create socket (wrap for ssl as needed)
	sock = socket.socket(family=url[1])
	if port == 443:
		sock = ssl.wrap_socket(sock, ssl_version=3)
	sock.settimeout(0.08)

	# Send request, and return "None" if anything goes wrong
	try:
		rtt = time.time()
		sock.connect((url[0], port))
		sock.send(b"HEAD / HTTP/1.1\r\n\r\n")
		ret = sock.recv(1000)
		rtt = time.time() - rtt
	except (OSError, ConnectionRefusedError, socket.gaierror, socket.timeout) as e:
		utils.error(Exception("Could not connect to %s: %s" % (url[0], e)))
		return None
	except ssl.SSLError as e:
		utils.warn("SSL handshake with %s failed: %s" % (url[0], e))
		return None
	finally:
		sock.close()

	# Servers that enforce ssl encryption when our socket isn't wrapped - or don't
	# recognize encrypted requests when it is - will sometimes send empty responses
	if not ret:
		return None

	# Check for "Server" header if available.
	# Note - this assumes that both the contents of the "Server" header and the response code are
	# utf8-decodable, which may need to be patched in the future
	try:
		srv = ret.index(b'Server: ')
	except ValueError:
		return rtt*1000, ret[9:12].decode(), "Unkown"
	return rtt*1000, ret[9:12].decode(), ret[srv+8:ret.index(b'\r', srv)].decode()


def mysql(url: utils.Host) -> typing.Optional[typing.Tuple[float, str]]:
	"""
	Checks for a MySQL server running on the host specified by url.
	Returns the server version if one is found, else None.
	"""

	sock = socket.socket(family=url[1])
	sock.settimeout(0.08)
	try:
		rtt = time.time()
		sock.connect((url[0], 3306))
		return (time.time() - rtt)* 1000, sock.recv(1000)[5:10].decode()
	except (UnicodeError, OSError, ConnectionRefusedError, socket.gaierror, socket.timeout) as e:
		utils.error(Exception("Could not connect to %s: %s" % (url[0], e)))
		return None
	finally:
		sock.close()

def portScan(host:utils.Host, pool:multiprocessing.pool.Pool)-> typing.Tuple[str, utils.ScanResult]:
	"""
	Scans a host using a multiprocessing worker pool to see if a specific set of ports are open,
	possibly returning extra information in the case that they are.

	Returns a tuple of (host, information) where host is the ip of the host scanned and information
	is any and all information gathered from each port as a tuple in the order (80, 443).
	If the specified port is not open, its spot in the tuple will contain `None`, but will otherwise
	contain some information related to the port.
	"""

	# Dispatch the workers
	hypertext = pool.apply_async(http, (host,))
	https = pool.apply_async(http, (host, 443))
	mysqlserver = pool.apply_async(mysql, (host,))

	# Collect and return
	return utils.ScanResult(hypertext.get(), https.get(), mysqlserver.get())



# Copyright 2018 Comcast Cable Communications Management, LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module defines a single function which implements route tracing.
"""
import socket
import struct
from struct import unpack
import time
from . import utils

class Tracer():
	"""
	A context-manage-able route tracer.
	"""

	def __init__(self, host:utils.Host, ID:int, maxHops:int):
		"""
		Constructs a route tracer, including the sockets used to send/recieve
		route tracer packets.

		Not that - unlike the functional implementation - this takes a direct
		integer instead of a whole `Config` object to determine `maxHops`.
		"""
		self.host = host
		self.ID = ID
		self.maxHops = maxHops

		# A bunch of stuff needs to be tweaked if we're using IPv6
		if host.family is socket.AF_INET6:
			# BTW, this doesn't actually work. The RFCs for IPv6 don't define
			# the behaviour of raw sockets - which are heavily utilized by
			# `connvitals`. One of these days, I'll have to implement it using
			# raw ethernet frames ...

			self.receiver = socket.socket(family=host.family,
			                              type=socket.SOCK_RAW,
			                              proto=socket.IPPROTO_ICMPV6)
			self.setTTL = self.setIPv6TTL
			self.isMyTraceResponse = self.isMyIPv6TraceResponse

		else:
			self.receiver = socket.socket(family=host.family,
			                              type=socket.SOCK_RAW,
			                              proto=socket.IPPROTO_ICMP)

		# We need a sender because a UDP socket can't receive ICMP 'TTL
		# Exceeded In Transit' packets, and having a raw sender introduces
		# a slew of new issues.
		self.sender = socket.socket(family=host.family,
		                            type=socket.SOCK_DGRAM)
		self.receiver.settimeout(0.05)

	def __enter__(self) -> 'Tracer':
		"""
		Context-managed instantiation.
		"""
		return self

	def __exit__(self, exc_type, exc_value, traceback):
		"""
		Context-managed cleanup.
		"""
		self.sender.shutdown(socket.SHUT_RDWR)
		self.sender.close()
		self.receiver.shutdown(socket.SHUT_RDWR)
		self.receiver.close()

		# Print exception information if possible
		if exc_type and exc_value:
			utils.error(exc_type("Unknown error occurred in route trace"))
			utils.error(exc_type(exc_value))
			if traceback:
				utils.warn("Stack Trace for route trace error: %s" % traceback)

	def __del__(self):
		"""
		Non-context-managed cleanup.
		"""
		try:
			self.sender.close()
			self.receiver.close()
		except AttributeError:
			# At least one of the socket references was already deleted
			pass

	def trace(self) -> utils.Trace:
		"""
		Runs the route trace, returning a list of visited hops
		"""
		from time import time
		ret = []

		for ttl in range(1, self.maxHops+1):
			self.setTTL(ttl)

			rtt = time()

			try:
				self.sender.sendto(b'', (self.host.addr, self.ID))
			except OSError:
				ret.append(utils.TraceStep("*", -1))

			try:
				while True:
					pkt, addr = self.receiver.recvfrom(1024)
					rtt = time() - rtt

					if self.isMyTraceResponse(pkt):
						break

			except socket.timeout:
				ret.append(utils.TraceStep("*", -1))
			else:
				ret.append(utils.TraceStep(addr[0], rtt*1000))
				if addr[0] == self.host.addr:
					break

		return utils.Trace(ret)

	def setIPv4TTL(self, ttl:int):
		"""
		Sets the TTL assuming `sender` is an IPV4 socket.
		"""
		self.sender.setsockopt(socket.SOL_IP, socket.IP_TTL, ttl)

	def setIPv6TTL(self, ttl:int):
		"""
		Sets the TTL assuming `sender` is an IPV6 socket.
		"""
		# Actually, hop limits should be set at the packet level, so this'll
		# change when I move to ethernet frames.
		self.sender.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_UNICAST_HOPS, ttl)

	def isMyIPv4TraceResponse(self, pkt:bytes) -> bool:
		"""
		Returns `True` if `pkt` is an IPv4 Traceroute Response AND it came
		from this particular Tracer - otherwise `False`.
		"""
		return pkt[20] in {11,3} and unpack("!H", pkt[50:52])[0] == self.ID

	def isMyIPv6TraceResponse(self, pkt:bytes) -> bool:
		"""
		Returns `True` if `pkt` is an IPv6 Traceroute Response AND it came
		from this particular Tracer - otherwise `False`
		"""
		return x[0] in {1,3} # ID fetch not implemented, since this isn't actually supported yet.

	# IPv4 is default
	setTTL = setIPv4TTL
	isMyTraceResponse = isMyIPv4TraceResponse


# Functional implementation still exists for convenience/legacy compatibility.

def trace(host: utils.Host, myID: int, config: 'config.Config') -> utils.Trace:
	"""
	Traces a route from the localhost to a given destination.
	Returns a tabular list of network hops up to the maximum specfied by 'hops'
	"""

	ret = []

	ipv6 = host[1] == socket.AF_INET6

	receiver = socket.socket(family=host.family, type=socket.SOCK_RAW, proto=58 if ipv6 else 1)
	receiver.settimeout(0.05)
	sender = socket.socket(family=host.family, type=socket.SOCK_DGRAM, proto=17)

	# Sets up functions used in the main loop, so it can transparently
	# handle ipv4 and ipv6 without needing to check which one we're
	# using on every iteration.
	setTTL, isTraceResponse, getIntendedDestination = None, None, None
	getID = lambda x: (x[50] << 8) + x[51]
	if ipv6:
		setTTL = lambda x: sender.setsockopt(41, 4, x)
		isTraceResponse = lambda x: x[0] in {1, 3}
		getIntendedDestination = lambda x: socket.inet_ntop(socket.AF_INET6, x[32:48])
	else:
		setTTL = lambda x: sender.setsockopt(socket.SOL_IP, socket.IP_TTL, x)
		isTraceResponse = lambda x: x[20] in {11, 3}
		getIntendedDestination = lambda x: ".".join(str(byte) for byte in x[44:48])

	for ttl in range(config.HOPS):
		setTTL(ttl+1)
		timestamp = time.time()

		try:
			sender.sendto(b'', (host[0], myID))
		except OSError as e:
			ret.append(utils.TraceStep("*", -1))
			continue

		try:
			#Wait for packets sent by this trace
			while True:
				pkt, addr = receiver.recvfrom(1024)
				rtt = time.time() - timestamp

				# If this is a response from a tracer and the tracer sent
				# it to the same place we're sending things, then this
				# packet must belong to us.
				if isTraceResponse(pkt):
					destination = getIntendedDestination(pkt)
					if destination == host.addr and getID(pkt) == myID:
						break

		except socket.timeout:
			ret.append(utils.TraceStep("*", -1)),
			done = False
		else:
			ret.append(utils.TraceStep(addr[0], rtt*1000))
			done = addr[0] == host[0]

		if done:
			break
	receiver.close()
	sender.close()
	return utils.Trace(ret)



# Copyright 2018 Comcast Cable Communications Management, LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A utility to check connection vitals with a remote host.

	Usage: connvitals [ -h --help ] [ -H --hops HOPS ] [ -p --pings PINGS ] [ -P --no-ping ]
	                  [ -t --trace ] [ --payload-size PAYLOAD ] [ --port-scan ] [ -j --json ]
	                  host [hosts... ]

Each 'host' can be an ipv4 address, ipv6 address, or a fully-qualified domain name.

Submodules:
	utils:      Contains utility functionality such as error/warning reporting and host address parsing
	ping:       Groups functionality related to ICMP/ICMPv6 tests
	traceroute: Contains a function for tracing a route to a host
	ports:      Specifies functions for checking specific host ports for http(s) and MySQL servers

"""

__version__ = "4.3.2"
__author__ = "Brennan Fieck"

def main() -> int:
	"""
	Runs the utility with the arguments specified on sys.argv.
	Returns: Always 0 to indicate "Success", unless the utility terminates
		prematurely with a fatal error.
	"""

	from . import config

	config.init()

	# No hosts could be parsed
	if not config.CONFIG or not config.CONFIG.HOSTS:
		from . import utils
		utils.error("No hosts could be parsed! Exiting...", True)

	from . import collector

	collectors = [collector.Collector(host,i+1) for i,host in enumerate(config.CONFIG.HOSTS)]

	# Start all the collectors
	for collect in collectors:
		collect.start()

	# Wait for every collector to finish
	# Print JSON if requested
	for collect in collectors:
		_ = collect.join()
		collect.result = collect.recv()
		print(repr(collect) if collect.conf.JSON else collect)

	# Errors will be indicated on stdout; because we query multiple hosts, as
	# long as the main routine doesn't crash, we have exited successfully.
	return 0



# Copyright 2018 Comcast Cable Communications Management, LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module defines the config options for the 'connvitals' command
"""

import socket
from . import __version__, utils

# Configuration values

class Config():
	"""
	Represents a configuration.
	"""

	def __init__(self,*,HOPS     = 30,
	                    JSON     = False,
	                    PAYLOAD  = b'The very model of a modern Major General.',
	                    TRACE    = False,
	                    NOPING   = False,
	                    PORTSCAN = False,
	                    NUMPINGS = 10,
	                    HOSTS    = None):
		"""
		Initializes a configuration
		"""
		self.HOPS     = HOPS
		self.JSON     = JSON
		self.PAYLOAD  = PAYLOAD
		self.TRACE    = TRACE
		self.NOPING   = NOPING
		self.PORTSCAN = PORTSCAN
		self.NUMPINGS = NUMPINGS
		self.HOSTS    = HOSTS if HOSTS is not None else {}

CONFIG = None

def init():
	"""
	Initializes the configuration.
	"""
	global __version__, CONFIG

	from argparse import ArgumentParser as Parser
	parser = Parser(description="A utility to check connection vitals with a remote host.",
	              epilog="'host' can be an ipv4 or ipv6 address, or a fully-qualified domain name.")

	parser.add_argument("hosts",
	                    help="The host or hosts to check connection to. "\
	                         "These can be ipv4 addresses, ipv6 addresses, fqdn's, "\
	                         "or any combination thereof.",
	                    nargs="*")

	parser.add_argument("-H", "--hops",
	                    dest="hops",
	                    help="Sets max hops for route tracing (default 30).",
	                    default=30,
	                    type=int)

	parser.add_argument("-p", "--pings",
	                    dest="numpings",
	                    help="Sets the number of pings to use for aggregate statistics (default 10).",
	                    default=10,
	                    type=int)

	parser.add_argument("-P", "--no-ping",
	                    dest="noping",
	                    help="Don't run ping tests.",
	                    action="store_true")

	parser.add_argument("-t", "--trace",
	                    dest="trace",
	                    help="Run route tracing.",
	                    action="store_true")

	parser.add_argument("-s", "--port-scan",
	                    dest="portscan",
	                    help="Scan the host(s)'s ports for commonly-used services",
	                    action="store_true")

	parser.add_argument("--payload-size",
	                    dest="payload",
	                    help="Sets the size (in B) of ping packet payloads (default 41).",
	                    default=b'The very model of a modern Major General.',
	                    type=int)

	parser.add_argument("-j", "--json",
	                    dest="json",
	                    help="Outputs in machine-readable JSON (no newlines)",
	                    action="store_true")

	parser.add_argument("-V", "--version",
	                    action="version",
	                    version="%(prog)s "+__version__)

	args = parser.parse_args()

	# Before doing anything else, make sure we have permission to open raw sockets
	try:
		sock = socket.socket(socket.AF_INET, socket.SOCK_RAW, proto=1)
		sock.close()
		del sock
	except PermissionError:
		from sys import argv
		utils.error(PermissionError("You do not have the permissions necessary to run %s" % (argv[0],)))
		utils.error("(Hint: try running as root, with `capsh` or with `sudo`)", True)

	CONFIG = Config(HOPS     = args.hops,
	                JSON     = args.json,
	                PAYLOAD  = args.payload,
	                TRACE    = args.trace,
	                NOPING   = args.noping,
	                PORTSCAN = args.portscan,
	                NUMPINGS = args.numpings)

	# Parse the list of hosts and try to find valid addresses for each
	CONFIG.HOSTS = {}

	for host in args.hosts:
		info = utils.getaddr(host)
		if not info:
			utils.error("Unable to resolve host ( %s )" % host)
		else:
			CONFIG.HOSTS[host] = info



#!/usr/bin/env python3
# Copyright 2018 Comcast Cable Communications Management, LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The setuptools-based installer for connvitals
"""

import os
import sys

# RPMs generated for fedora/rhel/centos need to have a different name
# (debian/ubuntu automatically prepends python3-, but those do not)
import platform
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

sys.path.append(here)
import connvitals

with open(os.path.join(here, 'README.rst')) as f:
	long_description = f.read()

setup(
	name="connvitals",
	version=connvitals.__version__,
	description='Checks a machines connection to a specific host or list of hosts',
	long_description=long_description,
	url='https://github.com/connvitals',
	author='Brennan Fieck',
	author_email='Brennan_WilliamFieck@comcast.com',
	classifiers=[
		'Development Status :: 5 - Production/Stable',
		'Intended Audience :: Telecommunications Industry',
		'Intended Audience :: Developers',
		'Intended Audience :: Information Technology',
		'Topic :: Internet',
		'Topic :: Internet :: WWW/HTTP',
		'Topic :: Scientific/Engineering :: Information Analysis',
		'Topic :: Utilities',
		'License :: OSI Approved :: Apache Software License',
		'Environment :: Console',
		'Operating System :: OS Independent',
		'Programming Language :: Python :: Implementation :: CPython',
		'Programming Language :: Python :: Implementation :: PyPy',
		'Programming Language :: Python :: 3 :: Only',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7'
	],
	keywords='network statistics connection ping traceroute port ip',
	packages=['connvitals'],
	install_requires=['setuptools', 'typing'],
	entry_points={
		'console_scripts': [
			'connvitals=connvitals.__init__:main',
		],
	},
	python_requires='~=3.4'
)



# Copyright 2018 Comcast Cable Communications Management, LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains utility functions used by the main utility to do things
lke printing errors and warnings to stderr, or get a single, valid IP address
for a host.
"""

import typing
import socket

# I don't know why, but pylint seems to think that socket.AddressFamily isn't real, but it is.
# Nobody else has this issue as far as I could find.
#pylint: disable=E1101
# Host = typing.NamedTuple("Host", [('addr', str), ('family', socket.AddressFamily), (name, str)])
Host = typing.NamedTuple("Host", [('addr', str), ('family', socket.AddressFamily)])
#pylint: enable=E1101

PingResult = typing.NamedTuple("PingResult", [
                                  ('minimum', float),
                                  ('avg',     float),
                                  ('maximum', float),
                                  ('std',     float),
                                  ('loss',    float)])

def pingResultToStr(self: PingResult) -> str:
	"""
	Returns the string representation of a ping result in plaintext
	"""
	fmt = "%.3f\t%.3f\t%.3f\t%.3f\t%.3f"
	return fmt % (self.minimum, self.avg, self.maximum, self.std, self.loss)

def pingResultRepr(self: PingResult) -> str:
	"""
	Returns the JSON representation of a ping result
	"""
	fmt = '{"min":%f,"avg":%f,"max":%f,"std":%f,"loss":%f}'
	return fmt % (self.minimum, self.avg, self.maximum, self.std, self.loss)

PingResult.__str__ = pingResultToStr
PingResult.__repr__ = pingResultRepr


TraceStep = typing.NamedTuple("TraceStep", [("host", str), ("rtt", float)])
Trace = typing.NewType("Trace", typing.List[TraceStep])

def traceStepToStr(self: TraceStep) -> str:
	"""
	Returns the string representation of a step of a route trace in plaintext

	>>> traceStepToStr(TraceStep("1.2.3.4", 3.059267))
	'1.2.3.4\\t3.059'
	>>> traceStepToStr(TraceStep("*", -1))
	'*'
	"""
	if self.rtt < 0 or self.host == "*":
		return "*"
	return "%s\t%.3f" % (self.host, self.rtt)

def traceStepRepr(self: TraceStep) -> str:
	"""
	Returns the JSON representation of a single step in a route trace

	>>> traceStepRepr(TraceStep("1.2.3.4", 3.059267))
	'["1.2.3.4", 3.059267]'
	>>> traceStepRepr(TraceStep("*", -1))
	'["*"]'
	"""
	if self.rtt < 0 or self.host == "*":
		return '["*"]'
	return '["%s", %f]' % (self.host, self.rtt)

def compareTraceSteps(self: TraceStep, other: TraceStep) -> bool:
	"""
	Implements `self == other`

	Two trace steps are considered equal iff their hosts are the same - rtt is not considered.

	>>> compareTraceSteps(TraceStep("localhost", -800), TraceStep("localhost", 900))
	True
	>>> compareTraceSteps(TraceStep("localhost", 7), TraceStep("127.0.0.1", 7))
	False
	"""
	return self.host == other.host

def traceStepIsValid(self: TraceStep) -> bool:
	"""
	Implements `bool(self)`

	Returns True if the step reports that the packet reached the host within the timeout,
	False otherwise.

	>>> traceStepIsValid(TraceStep('*', -1))
	False
	>>> traceStepIsValid(TraceStep("someaddr", 0))
	True
	>>> traceStepIsValid(TraceStep("someotheraddr", 27.0))
	True
	"""
	return self.rtt >= 0 and self.host != "*"

TraceStep.__str__ = traceStepToStr
TraceStep.__repr__ = traceStepRepr
TraceStep.__eq__ = compareTraceSteps
TraceStep.__bool__ = traceStepIsValid

def compareTraces(self: Trace, other: Trace) -> bool:
	"""
	Implements `self == other`

	Checks that traces are of the same length and contain the same hosts in the same order
	i.e. does *not* check the rtts of any or all trace steps.

	Note: ignores steps that are invalid ('*').

	>>> a = Trace([TraceStep('0.0.0.1', 0), TraceStep('0.0.0.2', 0)])
	>>> b=Trace([TraceStep('0.0.0.1',0), TraceStep('*',-1), TraceStep('*',-1), TraceStep('0.0.0.2',0)])
	>>> compareTraces(a, b)
	True
	"""
	this, that = [step for step in self if step], [step for step in other if step]
	return len(this) == len(that) and all(this[i] == that[i] for i in range(len(this)))

def traceToStr(self: Trace) -> str:
	"""
	Implements `str(self)`

	Returns the plaintext representation of a route trace.
	"""
	return '\n'.join(str(step) for step in self)

def traceRepr(self: Trace) -> str:
	"""
	Implements `repr(self)`

	Returns the JSON representation of a route trace.
	"""
	return "[%s]" % ','.join(repr(step) for step in self)

Trace.__str__ = traceToStr
Trace.__repr__ = traceRepr
Trace.__eq__ = compareTraces


ScanResult = typing.NamedTuple("ScanResult", [("httpresult", typing.Tuple[float, str, str]),
                                              ("httpsresult", typing.Tuple[float, str, str]),
                                              ("mysqlresult", typing.Tuple[float, str])])

def scanResultToStr(self: ScanResult) -> str:
	"""
	Returns the string representation of a portscan result in plaintext
	"""
	return "%s\t%s\t%s" % ("%.3f, %s, %s" % self.httpresult if self.httpresult else 'None',
	                       "%.3f, %s, %s" % self.httpsresult if self.httpsresult else 'None',
	                       "%.3f, %s" % self.mysqlresult if self.mysqlresult else 'None')

def scanResultRepr(self: ScanResult) -> str:
	"""
	Returns the JSON representation of a portscan result
	"""
	httpFmt = '{"rtt":%f,"response code":"%s","server":"%s"}'
	http = httpFmt % self.httpresult if self.httpresult else '"None"'
	https = httpFmt % self.httpsresult if self.httpsresult else '"None"'
	mySQL = '{"rtt":%f,"version":"%s"}' % self.mysqlresult if self.mysqlresult else '"None"'
	return '{"http":%s,"https":%s,"mysql":%s}' % (http, https, mySQL)

ScanResult.__str__ = scanResultToStr
ScanResult.__repr__ = scanResultRepr


def error(err: Exception, fatal: int=False):
	"""
	Logs an error to stderr, then exits if fatal is a non-falsy value, using it as an exit code
	"""
	from sys import stderr
	from time import ctime
	if stderr.isatty():
		fmt = "\033[38;2;255;0;0mEE: %s:"
		print(fmt % type(err).__name__, "%s" % err, "-\t", ctime(), "\033[m", file=stderr)
	else:
		print("EE: %s:" % type(err).__name__, "%s" % err, "-\t", ctime(), file=stderr)
	if fatal:
		exit(int(fatal))

def warn(warning: str):
	"""
	Logs a warning to stderr.
	"""
	from sys import stderr
	from time import ctime
	if stderr.isatty():
		print("\033[38;2;238;216;78mWW:", warning, "-\t", ctime(), "\033[m", file=stderr)
	else:
		print("WW:", warning, '-\t', ctime(), file=stderr)

def getaddr(host: str) -> typing.Optional[Host]:
	"""
	Returns a tuple of Address Family, IP Address for the host passed in `host`.
	"""

	try:
		addrinfo = socket.getaddrinfo(host, 1).pop()
		return Host(addrinfo[4][0], addrinfo[0])
	except socket.gaierror:
		return None
